//==============================================================================
// ContigEvalUtils.cc
//
// ----------------------------------------------------------------------------
// $Id: //CBT-depot/CompBioToolsApps/bac-stitcher/ContigEvalUtils.cc#11 $
//------------------------------------------------------------------------------
// Utilities to aid in finding problems in contigs.  
//
//==============================================================================
//.......1.........2.........3.........4.........5.........6.........7.........8

#include "ContigEvalUtils.hh"

int trimWindow = 20;
float trimHeight = 1.5;


void FindLeftTail(vector<int> &TCover, int &leftTail)
{
  // Find leftTail
  int TSize = TCover.size();
  for(int i = 0;i< TSize;i++){
	 int sum = 0;
	 for(int w = 0;w<trimWindow;w++){
		int idx = w + i;
		if (idx < TCover.size()){
		  sum+=TCover[idx];
		}
	 }
	 float avgCover = (float)sum/(float)trimWindow;
	 if (avgCover >=trimHeight){
		leftTail = i;
		break;
	 }
  }
}


void FindRightTail(vector<int> &TCover,int &rightTail)
{
  // Find rightTail
  int TSize = TCover.size();
  for(int i = 0;i< TSize;i++){

	 int sum = 0;
	 for(int w = 0;w<trimWindow;w++){
		int idx = TCover.size() - w - i -1;
		if (idx >=0){
		  sum+=TCover[idx];
		}
	 }
	 float avgCover = (float)sum/(float)trimWindow;
	 if (avgCover >=trimHeight){
		rightTail = TCover.size()-i-1;
		break;
	 }
  }
}

//---------------------------------------------------
// * FindTail
//---------------------------------------------------
//
void FindTail(vector<int> &TCover,int &leftTail, int &rightTail)
{
  leftTail = -1;
  rightTail = -1;

  //int halfTSize = (int) (TCover.size()/2.0);

  // Find leftTail
  int TSize = TCover.size();
  for(int i = 0;i< TSize;i++){
	 int sum = 0;
	 for(int w = 0;w<trimWindow;w++){
		int idx = w + i;
		if (idx < TCover.size()){
		  sum+=TCover[idx];
		}
	 }
	 float avgCover = (float)sum/(float)trimWindow;
	 if (avgCover >=trimHeight){
		leftTail = i;
		break;
	 }
  }

  // Find rightTail
  for(int i = 0;i< TSize;i++){

	 int sum = 0;
	 for(int w = 0;w<trimWindow;w++){
		int idx = TCover.size() - w - i -1;
		if (idx >=0){
		  sum+=TCover[idx];
		}
	 }

	 float avgCover = (float)sum/(float)trimWindow;

	 //if (gBugFlag) cout<<"\t"<<"i="<<i<<" sum="<<sum<<"\tavgCover="<<avgCover<<endl;

	 if (avgCover >=trimHeight){
		rightTail = TCover.size()-i-1;
		break;
	 }
  }
}


//---------------------------------------------------
// * FindCrash2
//---------------------------------------------------
// Like FindCrash, but skips ends that aren't trimmed. 
//
//
void FindCrash2(int leftTail,int rightTail, vector<int> &TCover,int &leftCrash, int &rightCrash)
{
  leftCrash = -1;
  rightCrash = -1;

 
  // Find leftCrash
  int TSize = TCover.size();
  for(int i = leftTail+20;i< rightTail-20;i++){
	 int sum = 0;
	 for(int w = 0;w<trimWindow;w++){		
		int idx = w + i;
		if (idx < TCover.size()){
		  sum+=TCover[idx];
		}
	 }
	 float avgCover = (float)sum/(float)trimWindow;
	 if (avgCover < trimHeight){
		leftCrash = i;
		break;
	 }
  }

  if (leftCrash < 20) {
	 leftCrash = -1;
	 return;
  }

  // Find rightCrash
  for(int i = leftCrash;i< rightTail;i++){

	 int sum = 0;
	 for(int w = 0;w<trimWindow;w++){
		int idx = w+i;
		if (idx >=0){
		  sum+=TCover[idx];
		}
	 }
	 float avgCover = (float)sum/(float)trimWindow;
	 if (avgCover >=trimHeight){
		rightCrash = i;
		break;
	 }
  }

  if (rightCrash >= (TSize -20)){
	 rightCrash = -1;
	 return;
  }
}


//---------------------------------------------------
// * FindCrash
//---------------------------------------------------
//
void FindCrash(vector<int> &TCover,int &leftTail, int &rightTail)
{
  leftTail = -1;
  rightTail = -1;
  
 // Find leftTail
  int TSize = TCover.size();
  for(int i = 0;i< TSize;i++){
	 int sum = 0;
	 for(int w = 0;w<trimWindow;w++){		
		int idx = w + i;
		if (idx < TCover.size()){
		  sum+=TCover[idx];
		}
	 }
	 float avgCover = (float)sum/(float)trimWindow;
	 if (avgCover < trimHeight){
		leftTail = i;
		break;
	 }
  }

  if (leftTail < 20) {
	 leftTail = -1;
	 return;
  }

  // Find rightTail
  for(int i = leftTail;i< TSize;i++){

	 int sum = 0;
	 for(int w = 0;w<trimWindow;w++){
		int idx = w+i;
		if (idx >=0){
		  sum+=TCover[idx];
		}
	 }
	 float avgCover = (float)sum/(float)trimWindow;
	 if (avgCover > trimHeight){
		rightTail = i;
		break;
	 }
  }

  if (rightTail >= (TSize -20)){
	 rightTail = -1;
	 return;
  }
}



//--------------------------------------
// * GetProblemContigs
//--------------------------------------
// 
void GetProblemContigs(string GraphFile,StringSet &ProblemContigs)
{
  cout<<"Reading problem contigs from "<<GraphFile<<"..."<<endl;
  ifstream fin;
  fin.open(GraphFile.c_str());
  do{
	 string line;
	 GetLine(fin,line);
	 if (Contains(line,"Rep_node:")){
		int idx0 = line.find(":");
		line = line.substr(idx0+1,line.size()-idx0-1);
		
		int idx1 = line.find("_");
		int idx2 = line.find("_",idx1+1);

		// If there is only one underscore. 
		if (idx2 < 0){
		  string contigName = line.substr(0,idx1);
		  StripWhiteSpace(contigName);
		  cout<<"Adding "<<contigName<<" to problem contigs list."<<endl;
		  ProblemContigs.Add(contigName);
		}
	 }
  }while(!fin.eof());
}

void ExpandFeature(vector<int> FCover,int level,int & left, int &right)
{
  for(left = left;left >= 0; left--){
	 if (FCover[left] > level){
		break;
	 }
  }

  for(right = right; right < FCover.size();right++){
	 if (FCover[right] > level){
		break;
	 }
  }
}

bool FeatureBetween(vector<int> &fVec,int crashmax,int start,int end,int &FeatureLocation)
{
  FeatureLocation = -1;

  // if ((start < 0) || (end < 0)) cerr<<"ERROR: FeatureBetween. start or end < 0"<<endl;
  //if ((start >= fVec.size()) || (end >= fVec.size())) cerr<<"ERROR: FeatureBetween start/end > size."<<endl;

  if (start < 0) start = 0;
  if (end < 0) end = fVec.size();
  if (start >= fVec.size()) start = 0;
  if (end >= fVec.size()) end = 0;

  for(int i = start;i<=end;i++){
	 if (fVec[i] <= crashmax){
		FeatureLocation = i;
		return(true);
	 }
  }
  return(false);
}

int 
ContigChecks::FindClosestMate(int cIdx,int minSize){
	int outIdx;
	ContigCheck &CIn = (*this)[cIdx];
	if (CIn.mbUsed) return(-1);
	
	int minGap = 999999;
	int minIdx = -1;
	for(int i = 0;i< size();i++){
		if (i == cIdx) continue;

		ContigCheck &B = (*this)[i];
		if (B.mbUsed) continue;
		
		if ((CIn.mContigName == B.mContigName) && (CIn.mSide == "left") && (B.mSide == "right")){
			int gap = abs(CIn.mStart - B.mEnd);
			if (gap < minGap){
				minGap = gap;
				minIdx = i;
			}
		}else if ((CIn.mContigName == B.mContigName) && (CIn.mSide == "right") && (B.mSide == "left")){
			int gap = abs(B.mStart - CIn.mEnd);
			if (gap < minGap){
				minGap = gap;
				minIdx = i;
			}
		}
	}
	
	int thesize = size();
	if (minIdx >= thesize){
		BUG();
		cout<<"ERROR: minIdx>=size() in FindClosestMate"<<endl;
	}

	// Only return the closest mate if it is closer than 2Kbp. 
	if (minIdx >= 0){
		ContigCheck &B = (*this)[minIdx];
		if (minGap < minSize){
			return(minIdx);
		}else{
			return(-1);
		}
	}
	return(-1);
}



void 
ContigChecks::MergePairs(int msize){
	
	for(int i = 0;i< size();i++) (*this)[i].mbUsed = false;
	
	ContigChecks CNew;
	
	// For each checkpoint, find the closest other check point. 
	for(int i = 0;i< size();i++){
		int mateIdx = FindClosestMate(i,msize);			
		if (mateIdx >=0){
			int thesize = size();
			if (mateIdx >= size()){
				cout<<"ERROR: "<<mateIdx<<">="<<size()<<endl;
				continue;
			}
			ContigCheck &C = (*this)[i];
			ContigCheck &Mate = (*this)[mateIdx];
			
			cout<<"Merging Checkpoints: "<<C.mContigName<<" "<<C.mSide<<" "<<C.mStart<<"-"<<C.mEnd<<"\t";
			cout<<Mate.mContigName<<" "<<Mate.mSide<<" "<<Mate.mStart<<"-"<<Mate.mEnd<<endl;
						
			ContigCheck MergedCheck;
			MergedCheck.mContigName = C.mContigName;
			MergedCheck.mSide = "merged";
			MergedCheck.mbMerged = true;
			MergedCheck.mbUsed = false;
			int midpoint;
			if (C.mSide == "left"){
				midpoint = (int) ((float)abs(C.mStart - Mate.mEnd)/(float)2.0);
				if (C.mStart < Mate.mEnd){
					MergedCheck.mStart = C.mStart;
					MergedCheck.mEnd = Mate.mEnd;
				}else{
					MergedCheck.mStart = Mate.mEnd;
					MergedCheck.mEnd = C.mStart;
				}
			}else{
				midpoint = (int) ((float)abs(Mate.mStart - C.mEnd)/(float)2.0);
				if (Mate.mStart < C.mEnd){
					MergedCheck.mStart = Mate.mStart;
					MergedCheck.mEnd = C.mEnd;
				}else{
					MergedCheck.mStart = C.mEnd;
					MergedCheck.mEnd = Mate.mStart;
				}
			}		
			MergedCheck.mPairSize = MergedCheck.mEnd - MergedCheck.mStart+1;
			//			cout<<"\tMerged = "<<MergedCheck.mStart<<"-"<<MergedCheck.mEnd<<endl;

			CNew.push_back(MergedCheck);				 
			C.mbUsed = true;
			Mate.mbUsed = true;
		}
	}
	
	// Now, copy only the non-merged pairs back...
	for(int i = 0;i< size();i++){
		if (!(*this)[i].mbUsed){
			CNew.push_back((*this)[i]);
		}
	}
	//cout<<"Size = "<<(*this).size()<<" CNew.size="<<CNew.size()<<endl;
	(*this) = CNew;
	//cout<<"Size = "<<(*this).size()<<" CNew.size="<<CNew.size()<<endl;
}



void IdentifySplits(Contig &C,ContigChecks &Checks,
										vector<int> &leftSplits,vector<int> &rightSplits)
{
  // Get just the problems from this contig. 
  ContigChecks CChecks;
  Checks.GetChecksFrom(C.mName,CChecks);

  vector<int> TCover;
  vector<int> Cover;
  vector<int> TSpan;
  vector<int> BACCover,WGSCover;
  vector<int> BADTSpan;

  C.GetTemplateCoverage(TCover);
  C.GetCoverage(Cover);
  C.GetTemplateSpan(TSpan);
  C.GetWGSBACCoverage(WGSCover,BACCover);
  //C.GetBACTSpan(BADTSpan);
  
	//	cout<<"Checks.size="<<CChecks.size()<<endl;
  for(int c = 0;c< CChecks.size();c++){
	 ContigCheck Check = CChecks[c];

	 //cout<<"Check Name="<<Check.mContigName<<endl;

	 // Choose the most accurate coordinate from each check point. 
	 //int checkCoord;
	 //if (Check.mbMerged) checkCoord = Check.mStart;
	 //else if (Check.mSide == "left") checkCoord = Check.mEnd;
	 //else checkCoord = Check.mStart;

	 int left,right;

	 // Tests 1-4 attempt to expand the split area based on a solid feature like a 
	 // template crash.  This may result in expanding the split to the end of the 
	 // contig, which is a special case. 

	 int featureCoord;
	 int featureLevel = 0;	 

	 cout<<C.mName<<"\t";
	 if (Check.mbMerged){
		cout<<"Merged";
	 }
	 cout<<"Checkpoint "<<Check.mStart<<":"<<Check.mEnd<<"  -->\t\t";

	 // 1. Template 0-crash at point -> expand. add left/right split. mark check as used. 
	 if (FeatureBetween(TCover,featureLevel,Check.mStart,Check.mEnd,featureCoord)){
		//	 if (TCover[checkCoord] == 0){
		left=right=featureCoord;
		ExpandFeature(TCover,featureLevel,left,right);
		cout<<"RULE 1 TEMPLATE0    "<<left<<":"<<right<<endl;
		leftSplits.push_back(left);
		rightSplits.push_back(right);
		Check.mbUsed = true;		
		continue;
	 }
 
	 // 2. BAC 0-crash. -> expand.  if > 2Kbp. add l/r split.  mark check as used. 
	 if (FeatureBetween(BACCover,featureLevel,Check.mStart,Check.mEnd,featureCoord)){
		//	 if (BACCover[checkCoord] == 0){
		left=right=featureCoord;
		ExpandFeature(BACCover,featureLevel,left,right);
		int csize = abs(right-left);
		if (csize >= 2000){
		  cout<<"RULE 2 BIG BAC CRASH "<<left<<":"<<right<<endl;
		  leftSplits.push_back(left);
		  rightSplits.push_back(right);
		  Check.mbUsed = true;		
		  continue;
		}
	 }

	 // 3. span 0-crash at point -> expand.  add left/right split.  mark check as used. 
	 if (FeatureBetween(TSpan,featureLevel,Check.mStart,Check.mEnd,featureCoord)){
		//	 if (TSpan[checkCoord] == 0){
		left=right=featureCoord;
		ExpandFeature(TSpan,featureLevel,left,right);
		cout<<"RULE 3 TSPAN0       "<<left<<":"<<right<<endl;
		leftSplits.push_back(left);
		rightSplits.push_back(right);
		Check.mbUsed = true;		
		continue;
	 }

	 // 4. BAC 0-crash and merged check -> expand.  add left/right split.  mark check as used.
	 // The idea here is that a small BAC crash by itself isn't informative, but when it 
	 // straddles a merged check point, it's significant. 
	 if (FeatureBetween(BACCover,featureLevel,Check.mStart,Check.mEnd,featureCoord)){
		//	 if (TSpan[checkCoord] == 0){
		if (Check.mbMerged){
		  left=right=featureCoord;
		  ExpandFeature(TSpan,featureLevel,left,right);
		  cout<<"RULE 4 BAC CRASH+MG "<<left<<":"<<right<<endl;
		  leftSplits.push_back(left);
		  rightSplits.push_back(right);
		  Check.mbUsed = true;		
		  continue;
		}
	 }
				 
	 // 5. Merged pair <= 2Kbp-> use start end as left/right split.  mark check as used. 
	 if (Check.mbMerged){
		if (Check.mPairSize <= 2000){
			leftSplits.push_back(Check.mStart);
			rightSplits.push_back(Check.mEnd);
			cout<<"RULE 5 MERGEDPAIR  "<<Check.mStart<<":"<<Check.mEnd<<endl;
			Check.mbUsed = true;
			continue;
		}		  		
	 }
	 
	 cout<<"RULE NONE"<<endl;

	 // 6. Merged pair > 2Kbp and BADTSpan >=2 -> use start end as left/right split.  mark check as used.
	 

	 // If the check point passes none of the tests, nothing is done. 
  }
}

